from __future__ import annotations
from .shared import (  # noqa: F401
    Account,
    MapData,
    StorageAccount,
    AuthKey,
    AuthKeyType,
    MTP,
    Storage,
    Serialize,
    TDesktop,
)
